/* IBM_PROLOG_BEGIN_TAG                                                   */
/* This is an automatically generated prolog.                             */
/*                                                                        */
/*                                                                        */
/*                                                                        */
/* IBM CONFIDENTIAL                                                       */
/*                                                                        */
/* OBJECT CODE ONLY SOURCE MATERIALS                                      */
/*                                                                        */
/*    COPYRIGHT International Business Machines Corp. 2005, 2007         */
/* All Rights Reserved                                                    */
/*                                                                        */
/* The source code for this program is not published or otherwise         */
/* divested of its trade secrets, irrespective of what has been           */
/* deposited with the U.S. Copyright Office.                              */
/*                                                                        */
/* IBM_PROLOG_END_TAG                                                     */

// ABOUT:
// This overload file allows a regular non-stereo application to be 
// displayed correctly for viewing with active or passive stereo devices.
// Depending on the Stereo configuration file, some render nodes will 
// produce the left image, and others the right. 

#include <malloc.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdlib.h>
#include <string.h>
// wire.h contains the SVN OpenGL enumerants (e.g. GLMULTMATRIXD etc.) and must be included
#include "wire.h"

#define UNSET       0
#define LEFT        1
#define RIGHT       2
#define CAVE_LEFT   3
#define CAVE_RIGHT  4
#define CAVE_UP     5

float protuberance = 0.0;

typedef struct
{
    int         view;
    double      eyeSeparation, protuberance, eyeDistance, offset;
    GLenum      currentMatrixMode;
    GLdouble    frustumLeft, frustumRight, frustumTop, frustumBottom, frustumNear, frustumFar, frustumWidth;
    int         frustumIsSet;
}StereoInfo;

StereoInfo  *myStereoInfo=NULL;
char        *myDisplayName=NULL;

// void getProtuberance()
// Read Protuberance value from environment variables and set.
void getProtuberance()
{
    char *envValue;
    envValue=getenv("SVN_STEREO_PROTUBERANCE");

    if (!envValue)
    {
        protuberance = 0.0;
        return;
    }

    protuberance = atof(envValue);

    if (protuberance<0 || protuberance >1)
    {
        fprintf(stderr,"SVN Stereo Warning: wrong protuberance value: %f\n",protuberance);
        protuberance = 0.0;
        return;
    }

    fprintf(stderr,"SVN Stereo: protuberance parameter set to %.2f\n",protuberance);
}

// int checkLine(char *line, char *name)
// Part of the parser for stereo configuration files
int checkLine(char *line, char *name)
{
    int i;
    while(isblank(*line))
        line++;
    while ( (*name)!='\0' && (*line) )
    {
        if ( *line != *name)
            return 0;
        name++;
        line++;
    }

    if (*line)
        return 1;
    else
        return 0;
}

int parseLine(char *line)
{
    if ( strstr(line," STEREO_LEFT") )
    {
        //printf("Stereo left!\n");
        myStereoInfo->view = LEFT;
    }
    if ( strstr(line," STEREO_RIGHT") )
    {
        //    printf("Stereo right!\n");
        myStereoInfo->view = RIGHT;
    }
    if ( strstr(line," CAVE_LEFT") )
    {
        //    printf("Stereo right!\n");
        myStereoInfo->view = CAVE_LEFT;
    }
    if ( strstr(line," CAVE_RIGHT") )
    {
        //    printf("Stereo right!\n");
        myStereoInfo->view = CAVE_RIGHT;
    }

    return 1;
}

// int parser()
// Simple parser for processing the Stereo configuration file
int parser()
{
    char *filename;
    FILE *f;
    char *line;
    size_t sz;

    filename=getenv("SVN_STEREO_FILE");
    if (!filename)
    {
        printf("SVN Stereo overload: could not read environment variable SVN_STEREO_FILE\n");
        return 0;
    }

    f = fopen(filename,"r");
    if (f==NULL)
    {
        printf("SVN Stereo overload: could not find file %s\n",filename);
        return 0;
    }

    line=NULL;
    while (getline(&line, &sz, f) > 0)
    {
        if (checkLine(line,myDisplayName))
            if ( parseLine(line) )
                return 1;
            else
                return 0;
        free(line);
        line=NULL;
    }
    return 0;
}

// Storage for the original 'true' OpenGL functions
static glOp pTbl[NUM_GL_FUNCTIONS];

static void
_popTransformation();
static void
_pushStereoProjectionTransformation();

// Overloaded versions of several OpenGL functions required to create a stereo view
// 

static void
_glLoadIdentity()
{
    void(*glop)() = (void(*)()) pTbl[GLLOADIDENTITY];

    _popTransformation();

    (*glop)();

    if (myStereoInfo->currentMatrixMode == GL_PROJECTION)
        _pushStereoProjectionTransformation();
}

static void
_glLoadMatrixd(GLdouble *m)
{
    void (*lm)(GLdouble *) = (void(*)(GLdouble *)) pTbl[GLLOADMATRIXD];

    _popTransformation();

    (*lm)(m);

    if (myStereoInfo->currentMatrixMode == GL_PROJECTION)
        _pushStereoProjectionTransformation();

}

static void
_glLoadMatrixf(GLfloat *m)
{
    void (*lm)(GLfloat *) = (void(*)(GLfloat *)) pTbl[GLLOADMATRIXF];


    _popTransformation();

    (*lm)(m);

    if (myStereoInfo->currentMatrixMode == GL_PROJECTION)
        _pushStereoProjectionTransformation();
}

static void
_glMultMatrixd(GLdouble *m)
{
    void (*mm)(GLdouble *) = (void(*)(GLdouble *)) pTbl[GLMULTMATRIXD];

    _popTransformation();

    (*mm)(m);

    if (myStereoInfo->currentMatrixMode == GL_PROJECTION)
        _pushStereoProjectionTransformation();
}

static void
_glMultMatrixf(GLfloat *m)
{
    void (*mm)(GLfloat *) = (void(*)(GLfloat *)) pTbl[GLMULTMATRIXF];

    _popTransformation();

    (*mm)(m);

    if (myStereoInfo->currentMatrixMode == GL_PROJECTION)
        _pushStereoProjectionTransformation();
}


static void
_glScalef(GLfloat x, GLfloat y, GLfloat z)
{
    void (*lm)(GLfloat, GLfloat, GLfloat) = 
        (void(*)(GLfloat, GLfloat, GLfloat)) pTbl[GLSCALEF];

    _popTransformation();

    (*lm)(x, y, z);

    if (myStereoInfo->currentMatrixMode == GL_PROJECTION)
        _pushStereoProjectionTransformation();

}


static void
_glScaled(GLdouble x, GLdouble y, GLdouble z)
{
    void (*lm)(GLdouble, GLdouble, GLdouble) = 
        (void(*)(GLdouble, GLdouble, GLdouble)) pTbl[GLSCALED];

    _popTransformation();

    (*lm)(x, y, z);

    if (myStereoInfo->currentMatrixMode == GL_PROJECTION)
        _pushStereoProjectionTransformation();
}

static void
_glRotated(GLdouble a, GLdouble x, GLdouble y, GLdouble z)
{
    void (*lm)(GLdouble, GLdouble, GLdouble, GLdouble) = 
        (void(*)(GLdouble, GLdouble, GLdouble, GLdouble)) pTbl[GLROTATED];

    _popTransformation();

    (*lm)(a, x, y, z);

    if (myStereoInfo->currentMatrixMode == GL_PROJECTION)
        _pushStereoProjectionTransformation();

}

static void
_glRotatef(GLfloat a, GLfloat x, GLfloat y, GLfloat z)
{
    void (*lm)(GLfloat, GLfloat, GLfloat, GLfloat) = 
        (void(*)(GLfloat, GLfloat, GLfloat, GLfloat)) pTbl[GLROTATEF];

    _popTransformation();

    (*lm)(a, x, y, z);

    if (myStereoInfo->currentMatrixMode == GL_PROJECTION)
        _pushStereoProjectionTransformation();
}


static void
_glTranslated(GLdouble x, GLdouble y, GLdouble z)
{
    void (*lm)(GLdouble, GLdouble, GLdouble) = 
        (void(*)(GLdouble, GLdouble, GLdouble)) pTbl[GLTRANSLATED];

    _popTransformation();

    (*lm)(x, y, z);

    if (myStereoInfo->currentMatrixMode == GL_PROJECTION)
        _pushStereoProjectionTransformation();
}

static void
_glPushMatrix()
{
    void (*lm)() = (void(*)()) pTbl[GLPUSHMATRIX];

    _popTransformation();

    (*lm)();

    if (myStereoInfo->currentMatrixMode == GL_PROJECTION)
        _pushStereoProjectionTransformation();
}

static void
_glPopMatrix()
{
    void (*lm)() = (void(*)()) pTbl[GLPOPMATRIX];

    _popTransformation();

    (*lm)();

    if (myStereoInfo->currentMatrixMode == GL_PROJECTION)
        _pushStereoProjectionTransformation();
}

static void
_glFrustum(GLdouble l, GLdouble r, GLdouble b, GLdouble t, GLdouble n, GLdouble f)
{
    void(*frustum)(GLdouble, GLdouble, GLdouble, GLdouble, GLdouble, GLdouble) =
        (void(*)(GLdouble, GLdouble, GLdouble, GLdouble, GLdouble, GLdouble))pTbl[GLFRUSTUM];

    _popTransformation();
    (*frustum)(l,r,b,t,n,f);

    if (myStereoInfo->currentMatrixMode == GL_PROJECTION)
        _pushStereoProjectionTransformation();
}

static void
_glOrtho(GLdouble l, GLdouble r, GLdouble b, GLdouble t, GLdouble n, GLdouble f)
{
    void(* ortho)(GLdouble, GLdouble, GLdouble, GLdouble, GLdouble, GLdouble) =
        (void(*)(GLdouble, GLdouble, GLdouble, GLdouble, GLdouble, GLdouble))pTbl[GLORTHO];

    _popTransformation();

    (*ortho)(l, r, b, t, n, f);

    if (myStereoInfo->currentMatrixMode == GL_PROJECTION)
        _pushStereoProjectionTransformation();
}


static void
_glMatrixMode(GLenum mode)
{
    void(*glop)(GLenum) = (void(*)(GLenum)) pTbl[GLMATRIXMODE];

    (*glop)(mode);
    myStereoInfo->currentMatrixMode = mode;

}


static void
_glTranslatef(GLfloat x, GLfloat y, GLfloat z)
{
    void (*glop)(GLfloat, GLfloat, GLfloat) = 
        (void(*)(GLfloat, GLfloat, GLfloat)) pTbl[GLTRANSLATEF];

    _popTransformation();

    (*glop)(x,y,z);

    if (myStereoInfo->currentMatrixMode == GL_PROJECTION)
        _pushStereoProjectionTransformation();

}

#ifdef DVCOMMCHAR
static void
_DVCommChar(char c)
{

    void (*dvop)(char) = 
        (void(*)(char)) pTbl[DVCOMMCHAR];

    if (c=='+' && protuberance<0.9)
    {
        protuberance += 0.1;
        fprintf(stderr,"SVN Stereo overload: protuberance parameter set to %.2f\n",protuberance);
    }
    else if (c=='-' && protuberance>0.1)
    {
        protuberance-=0.1;
        fprintf(stderr,"SVN Stereo overload: protuberance parameter set to %.2f\n",protuberance);
    }
    else
        (*dvop)(c);
}
#endif

// void Overload(glOp *glOperationsTable, glOp *glSystemOperationsTable)
// Perform setup and the actual function replacement operations required
void Overload(glOp *glOperationsTable, glOp *glSystemOperationsTable)
{
    // Acquire the display hostname ("client" for the application node)
    char*(*_getDisplayName)() = 
        (char*(*)()) glOperationsTable[DVGETDISPLAYNAME];
    myDisplayName = _getDisplayName();

    if (myDisplayName==NULL)
    {
        fprintf(stderr,"Stereo overload: unable to get display name from system info, aborting overload...\n");
        return;
    }

    // Read in the stereo configuration file
    if (myStereoInfo==NULL)
        myStereoInfo = (StereoInfo*)malloc(sizeof(StereoInfo));
    myStereoInfo->view=UNSET;
    parser();

    switch(myStereoInfo->view)
    {
    case LEFT:
        fprintf(stderr,"--SVN renderer on display %s has viewpoint STEREO_LEFT\n",myDisplayName);
        break;
    case RIGHT:
        fprintf(stderr,"--SVN renderer on display %s has viewpoint STEREO_RIGHT\n",myDisplayName);
        break;
    case CAVE_LEFT:
        fprintf(stderr,"--SVN renderer on display %s has viewpoint CAVE_LEFT\n",myDisplayName);
        break;
    case CAVE_RIGHT:
        fprintf(stderr,"--SVN renderer on display %s has viewpoint CAVE_RIGHT\n",myDisplayName);
        break;
    case UNSET:
        fprintf(stderr,"--SVN renderer on display %s: no viewpoint found in config file, aborting stereo overloads...\n",myDisplayName);
        break;
    }

    if (myStereoInfo->view == UNSET)
        return;

    getProtuberance();


    // Perform the actual overloading, by first backing up the function pointers to pTbl
    // Then replacing the OpenGL functions with our own overloaded versions
    memcpy(pTbl, glOperationsTable, sizeof(pTbl));

    glOperationsTable[GLLOADIDENTITY] = (glOp)_glLoadIdentity;

    glOperationsTable[GLMATRIXMODE] = (glOp)_glMatrixMode;

    glOperationsTable[GLLOADMATRIXD] = (glOp)_glLoadMatrixd;
    glOperationsTable[GLLOADMATRIXF] = (glOp)_glLoadMatrixf;

    glOperationsTable[GLMULTMATRIXD] = (glOp)_glMultMatrixd;
    glOperationsTable[GLMULTMATRIXF] = (glOp)_glMultMatrixf;

    glOperationsTable[GLSCALEF] = (glOp)_glScalef;
    glOperationsTable[GLSCALED] = (glOp)_glScaled;

    glOperationsTable[GLTRANSLATEF] = (glOp)_glTranslatef;
    glOperationsTable[GLTRANSLATED] = (glOp)_glTranslated;

    glOperationsTable[GLROTATEF] = (glOp)_glRotatef;
    glOperationsTable[GLROTATED] = (glOp)_glRotated;

    glOperationsTable[GLPUSHMATRIX] = (glOp)_glPushMatrix;

    glOperationsTable[GLPOPMATRIX] = (glOp)_glPopMatrix;

    glOperationsTable[GLFRUSTUM] = (glOp)_glFrustum;

    glOperationsTable[GLORTHO] = (glOp)_glOrtho;  

#ifdef DVCOMMCHAR
    glOperationsTable[DVCOMMCHAR] = (glOp)_DVCommChar;
#endif

}

// Undo the stereo transformation
void _popTransformation()
{
    void(* glpop)() = (void(*)()) pTbl[GLPOPMATRIX];

    if (myStereoInfo->currentMatrixMode == GL_PROJECTION)
        (*glpop)();
}

// void getFrustumFromMatrix(GLdouble *m,StereoInfo *si)
// Retrieve the StereoInfo based on the current OpenGL Projection Matrix
void getFrustumFromMatrix(GLdouble *m,StereoInfo *si)
{
    double alfa,beta;

    alfa=m[10];
    beta=m[14];

    if (alfa==1 || alfa==-1)
    {
        si->frustumIsSet=0;
        return; //this is not a valid projection matrix
    }

    if (m[3]==0 && m[7]==0 && m[11]==0)
    {
        si->frustumIsSet=0;
        return; //this is not a perspective projection matrix
    }

    si->protuberance = protuberance; //default value
    si->frustumNear = beta / (alfa-1);
    si->frustumFar = beta / (alfa+1);
    si->frustumWidth = 2*si->frustumNear / m[0];  

    if (si->protuberance > 1.0e-3 )
        si->eyeDistance = si->frustumNear + (si->frustumFar - si->frustumNear) * si->protuberance;
    else
        si->eyeDistance = si->frustumNear;
    si->eyeSeparation = si->eyeDistance / 30;
    si->offset = si->eyeSeparation * (si->frustumNear / si->eyeDistance) / si->frustumWidth;

    si->frustumIsSet=1;
}

void
_pushStereoProjectionTransformation()
{
    GLdouble *currentMatrix = (GLdouble*)malloc(16*sizeof(GLdouble));

    // Acquire functions from our overloaded set.
    void(*glpush)() = 
        (void(*)()) pTbl[GLPUSHMATRIX];
    void (*getdv)(GLenum, GLdouble *) = 
        (void (*)(GLenum, GLdouble *)) pTbl[GLGETDOUBLEV];
    void (*ldm)(GLdouble *) = 
        (void(*)(GLdouble *)) pTbl[GLLOADMATRIXD];
    void(*gltranslate)(GLfloat, GLfloat, GLfloat) = 
        (void(*)(GLfloat, GLfloat, GLfloat)) pTbl[GLTRANSLATEF];
    void(*glrotate)(GLfloat, GLfloat, GLfloat, GLfloat) = 
        (void(*)(GLfloat, GLfloat, GLfloat, GLfloat)) pTbl[GLROTATEF];


    (*glpush)();

    (*getdv)(GL_PROJECTION_MATRIX, currentMatrix);

    getFrustumFromMatrix(currentMatrix, myStereoInfo);

    if (!myStereoInfo->frustumIsSet)
    {
        free(currentMatrix);
        return;
    }

    // Shift the camera position based on eye separation
    if (myStereoInfo->view==LEFT)
    {
        (*gltranslate)( (myStereoInfo->eyeSeparation)/2 , 0 , 0);
        (*glrotate)(2.0,0.0,-1.0,0.0);
    }
    else if (myStereoInfo->view==RIGHT)
    {
        (*gltranslate)( -(myStereoInfo->eyeSeparation)/2 , 0 , 0);
        (*glrotate)(2.0,0.0,1.0,0.0);
    }
    
    // Warning: CAVE options have not been tested.
    // Frustum culling issues could come into effect
    else if (myStereoInfo->view==CAVE_LEFT)
        (*glrotate)(-15,0,1,0);
    else if (myStereoInfo->view==CAVE_RIGHT)
        (*glrotate)(15,0,1,0);


    free(currentMatrix);
}

